# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import base64
import io
import logging
import os
import zlib
from typing import Any
from typing import Dict
from typing import Optional

import earthkit.data as ekd
import mir
from anemoi.inference.grib.templates import IndexTemplateProvider
from anemoi.inference.grib.templates import TemplateProvider

LOG = logging.getLogger(__name__)


class BaseTemplateProvider(IndexTemplateProvider):
    def load_template(self, grib: str, lookup: Dict[str, Any]) -> bytes:  # type: ignore
        return zlib.decompress(base64.b64decode(grib))


class MirTemplatesProvider(TemplateProvider):
    """Template provider using mir to make new grid templates."""

    def __init__(self, manager: Any, path: Optional[str] = None) -> None:
        """Initialise the MirTemplatesProvider instance.

        Parameters
        ----------
        manager : Any
            The manager instance.
        path : str
            The path to the base handles file.
            Expected to be in the templates yaml format, and
            have no grid / area keys.
        """
        self.manager = manager

        if path is None:
            path = os.path.join(os.path.dirname(__file__), "base_handles.yaml")

        self._base_template_provider = BaseTemplateProvider(manager, path)

    def _regrid_with_mir(self, base_template: bytes, grid: str, area: Optional[str] = None) -> bytes:
        """Regrid the base template using mir.

        Parameters
        ----------
        base_template : bytes
            Base grib handle template as bytes to regrid.
        grid : str
            Target grid for regridding.
        area : Optional[str], optional
            Target area for regridding, by default None

        Returns
        -------
        bytes
            Grib handle in bytes regridded.
        """

        mir_input = mir.GribMemoryInput(base_template)
        job_args = {"grid": grid}
        if area:
            job_args["area"] = area

        job = mir.Job(**job_args)
        buffer = io.BytesIO()

        job.execute(mir_input, buffer)

        return buffer.getvalue()

    def template(self, variable: str, lookup: Dict[str, Any]) -> ekd.Field:
        """Get the template for the given variable and lookup.

        Parameters
        ----------
        variable : str
            The variable to get the template for.
        lookup : Dict[str, Any]
            The lookup dictionary.

        Returns
        -------
        ekd.Field
            The template field.
        """
        _lookup = lookup.copy()

        grid = _lookup.pop("grid")
        area = _lookup.pop("area", None)

        if isinstance(grid, (list, tuple)):
            grid = "/".join(map(str, grid))
        if isinstance(area, (list, tuple)):
            area = "/".join(map(str, area))

        base_template = self._base_template_provider.template(variable, lookup)
        if base_template is None:
            raise ValueError(f"Base template not found for variable {variable} with lookup {lookup}")

        regridded_template = self._regrid_with_mir(base_template, str(grid), area)

        if len(regridded_template) == 0:
            raise ValueError(f"Regridded template is empty for variable {variable} with lookup {lookup}")

        return ekd.from_source("memory", regridded_template)[0]  # type: ignore



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any

import earthkit.data as ekd
from anemoi.inference.processor import Processor
from anemoi.transform.filters.orog_to_z import Orography

LOG = logging.getLogger(__name__)


class InferenceOrography(Orography):
    """Inference Orography Filter.

    Patches the data request to replace `z` with `orog` if `z` is present.
    """

    def patch_data_request(self, data_request: Any) -> Any:
        param = data_request.get("param")
        if param is None:
            return data_request

        if self.z in param and (data_request.get("levtype", "") == "pl" or data_request.get("levelist", [])):
            data_request["param"] = [self.orog if p == self.z else p for p in param]
        return data_request


class OrographyProcessor(Processor):
    """A processor that applies the InferenceOrography filter to the given fields."""

    def __init__(self, context: Any, **kwargs: Any) -> None:
        """Initialize the OrographyProcessor.

        Parameters
        ----------
        context : object
            The context in which the filter is being used.
        **kwargs : dict
            Additional keyword arguments to pass to the filter.
        """
        super().__init__(context)
        self.filter = InferenceOrography(**kwargs)

    def process(self, fields: ekd.FieldList) -> ekd.FieldList:
        """Process the given fields using the InferenceOrography filter.

        Parameters
        ----------
        fields : ekd.FieldList
            The fields to be processed.

        Returns
        -------
        ekd.FieldList
            The processed fields.
        """
        return self.filter.forward(fields)

    def patch_data_request(self, data_request: Any) -> Any:
        """Patch the data request using the filter.

        Parameters
        ----------
        data_request : object
            The data request to be patched.

        Returns
        -------
        object
            The patched data request.
        """
        return self.filter.patch_data_request(data_request)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from anemoi.inference.grib.templates import create_template_provider
from anemoi.plugins.ecmwf.inference.mir_templates import MirTemplatesProvider


def test_registration():
    template_provider = create_template_provider(None, "mir")  # type: ignore
    assert isinstance(
        template_provider, MirTemplatesProvider
    ), "Template provider should be an instance of MirTemplatesProvider"


if __name__ == "__main__":
    test_registration()



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from dataclasses import asdict
from dataclasses import dataclass
from datetime import timedelta
from typing import Any
from typing import Optional

import multio
from anemoi.inference.context import Context
from anemoi.inference.decorators import main_argument
from anemoi.inference.output import Output
from anemoi.inference.types import State
from anemoi.utils.grib import shortname_to_paramid

CONVERT_PARAM_TO_PARAMID = True

NULL_TO_REMOVE = "NULL_TO_REMOVE"


class BaseMetadata:
    def to_dict(self) -> dict[str, Any]:
        """Convert the dataclass to a dictionary representation.

        Will remove any fields that have the value NULL_TO_REMOVE.
        """
        dict_repr = asdict(self)
        # Remove None values
        dict_repr = {key: value for key, value in dict_repr.items() if value is not NULL_TO_REMOVE}
        return dict_repr


@dataclass
class UserDefinedMetadata(BaseMetadata):
    stream: str
    """Stream name, e.g. oper, enfo"""
    type: str
    """Type name, e.g. fc, an"""
    klass: str
    """Class name, e.g. od, ai, ..."""
    expver: str
    """Experiment version, e.g. 0001"""
    model: str
    """Model name, e.g. aifs-single, ..."""
    number: Optional[int] = NULL_TO_REMOVE
    """Ensemble number, e.g. 0,1,2"""
    numberOfForecastsInEnsemble: Optional[int] = NULL_TO_REMOVE
    """Number of ensembles in the forecast, e.g. 50"""

    def to_dict(self) -> dict[str, Any]:
        dict_repr = super().to_dict()
        dict_repr["class"] = dict_repr.pop("klass")
        if "numberOfForecastsInEnsemble" in dict_repr:
            dict_repr["misc-numberOfForecastsInEnsemble"] = dict_repr.pop("numberOfForecastsInEnsemble")
        return dict_repr

    def __post_init__(self):
        if isinstance(self.number, int) and not isinstance(self.numberOfForecastsInEnsemble, int):
            error_msg = "numberOfForecastsInEnsemble must be an integer if number is provided"
            raise AttributeError(error_msg)


@dataclass
class MultioMetadata(BaseMetadata):
    param: int
    """Param ID, e.g. 130"""
    levtype: str
    """Level type, e.g. sfc,pl,soil"""
    date: int
    """Reference date, e.g. 20220101"""
    time: int
    """Reference time, e.g. 1200"""
    step: int
    """Forecast step, e.g. 0,6,12,24"""
    grid: str
    """Grid name, e.g. n320, o96"""
    levelist: Optional[int] = NULL_TO_REMOVE
    """Level, e.g. 0,50,100"""

    timespan: Optional[str] = NULL_TO_REMOVE
    """Time span, e.g."""

    origin: Optional[str] = NULL_TO_REMOVE
    """Origin name, e.g. ecmf, ukmo"""
    packing: Optional[str] = "ccsds"
    """Packing type, e.g. ccsds"""
    repres: Optional[str] = None
    """Representation type"""

    def __post_init__(self):
        if self.repres is None:
            if any(self.grid.upper().startswith(prefix) for prefix in ["N", "O"]):
                self.repres = "gg"
            else:
                self.repres = "ll"


class MultioOutputPlugin(Output):

    api_version = "1.0.0"
    schema = None

    _server: multio.Multio = None

    def __init__(
        self,
        context: Context,
        plan: str | dict | multio.plans.Plan,
        *,
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
        **metadata: Any,
    ) -> None:
        super().__init__(
            context,
            output_frequency=output_frequency,
            write_initial_state=write_initial_state,
        )
        self._plan = plan

        try:
            self._user_defined_metadata = UserDefinedMetadata(**metadata)
        except TypeError as e:
            raise TypeError(f"Invalid metadata: {e}") from e

    def open(self, state: State) -> None:
        if self._server is None:
            with multio.MultioPlan(self._plan):
                self._server = multio.Multio()
        self._server.open_connections()
        self._server.write_parametrization(self._user_defined_metadata.to_dict())

    def write_initial_state(self, state: State) -> None:
        """Write the initial step of the state.

        Parameters
        ----------
        state : State
            The state object.
        """

        state = state.copy()

        self.reference_date = state["date"]
        state.setdefault("step", timedelta(0))

        return self.write_step(state)

    def write_step(self, state: State) -> None:
        """Write a step of the state with multio."""
        if self._server is None:
            raise RuntimeError("Multio server is not open, call `.open()` first.")

        reference_date = self.reference_date or self.context.reference_date
        step = state["step"]

        shared_metadata = {
            "step": int(step.total_seconds() // 3600),
            "grid": str(self.context.checkpoint.grid).upper(),
            "date": int(reference_date.strftime("%Y%m%d")),
            "time": reference_date.hour * 100,
        }

        for param, field in state["fields"].items():
            variable = self.checkpoint.typed_variables[param]
            if variable.is_computed_forcing:
                continue

            param = variable.grib_keys.get("param", param)
            if CONVERT_PARAM_TO_PARAMID:
                param = shortname_to_paramid(param)

            metadata = MultioMetadata(
                param=param,
                levtype=variable.grib_keys["levtype"],
                levelist=variable.level if not variable.is_surface_level else NULL_TO_REMOVE,
                **shared_metadata,
            )

            # Copy the field to ensure it is contiguous
            # Removes ValueError: ndarray is not C-contiguous
            self._server.write_field({**metadata.to_dict()}, field.copy(order="C"))

        self._server.flush()

    def close(self) -> None:
        if self._server is None:
            raise RuntimeError("Multio server is not open to close, call `.open()` first.")

        self._server.close_connections()
        self._server = None


@main_argument("path")
class MultioOutputGribPlugin(MultioOutputPlugin):
    """Multio output plugin for GRIB files.

    This plugin uses the multio library to write GRIB files.
    It is a subclass of the MultioOutputPlugin class.
    """

    def __init__(
        self, context: Context, path: str, append: bool = True, per_server: bool = False, **kwargs: Any
    ) -> None:
        """Multio Grib Output Plugin.

        Parameters
        ----------
        context : Context
            Model Runner
        path : str
            Path to write to
        append : bool
            Whether to append to the file or not
        per_server : bool
            Whether to write to a separate file per server or not
        """
        plan = multio.plans.Client(
            plans=[
                multio.plans.Plan(
                    name="output-to-file",
                    actions=[
                        multio.plans.EncodeMTG(geo_from_atlas=True),
                        multio.plans.Sink(
                            sinks=[
                                multio.plans.sinks.File(
                                    append=append,
                                    per_server=per_server,
                                    path=path,
                                )
                            ]
                        ),
                    ],
                )
            ]
        )
        super().__init__(context, plan=plan, **kwargs)


@main_argument("fdb_config")
class MultioOutputFDBPlugin(MultioOutputPlugin):
    """Multio output plugin to write to FDB.

    This plugin uses the multio library to write to FDB.
    It is a subclass of the MultioOutputPlugin class.
    """

    def __init__(self, context: Context, fdb_config: str, **kwargs: Any) -> None:
        """Multio FDB Output Plugin.

        Parameters
        ----------
        context : Context
            Model Runner
        fdb_config : str
            FDB Configuration file
        """
        plan = multio.plans.Client(
            plans=[
                multio.plans.Plan(
                    name="output-to-fdb",
                    actions=[
                        multio.plans.EncodeMTG(geo_from_atlas=True),
                        multio.plans.Sink(
                            sinks=[
                                multio.plans.sinks.FDB(
                                    config=fdb_config,
                                )
                            ]
                        ),
                    ],
                )
            ]
        )

        super().__init__(context, plan=plan, **kwargs)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from pathlib import Path

from anemoi.inference.inputs import create_input
from anemoi.inference.runners import create_runner
from anemoi.inference.testing import fake_checkpoints
from anemoi.inference.testing.mock_checkpoint import MockRunConfiguration


@fake_checkpoints
def test_plugin():
    config = MockRunConfiguration.load(
        (Path(__file__).parent / "configs/simple.yaml").absolute(),
        overrides=dict(input="polytope"),
    )
    runner = create_runner(config)
    input = create_input(runner, config.input)
    assert input is not None


if __name__ == "__main__":
    test_plugin()



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import Optional
from typing import Union

import earthkit.data as ekd
from anemoi.inference.context import Context
from anemoi.inference.grib.templates import TemplateProvider
from anemoi.inference.grib.templates import create_template_provider
from anemoi.inference.inputs.grib import GribInput
from anemoi.inference.types import DataRequest
from anemoi.inference.types import Date
from anemoi.inference.types import ProcessorConfig
from anemoi.inference.types import State
from anemoi.utils.grib import shortname_to_paramid
from earthkit.data.utils.dates import to_datetime

from .geopotential_height import OrographyProcessor

LOG = logging.getLogger(__name__)

SOIL_MAPPING = {"stl1": "sot", "stl2": "sot", "stl3": "sot", "swvl1": "vsw", "swvl2": "vsw", "swvl3": "vsw"}


def _retrieve_soil(request: dict, soil_params: list[str]) -> ekd.FieldList:
    """Retrieve soil data.

    Map the soil parameters to the correct ECMWF parameter IDs and levels.

    Parameters
    ----------
    request : dict
        Request for the soil data.
    soil_params : list[str]
        Parameters to be retrieved.

    Returns
    -------
    ekd.FieldList
        Soil data.
    """
    request = request.copy()

    request["param"] = list(SOIL_MAPPING[s] for s in soil_params)
    request["levelist"] = list(set(int(s[-1]) for s in soil_params))
    request.pop("levtype")

    soil_data = ekd.from_source("ecmwf-open-data", request)
    for field in soil_data:
        newname = {f"{v}{k[-1]}": k for k, v in SOIL_MAPPING.items()}[
            f"{field.metadata()['param']}{field.metadata()['level']}"
        ]
        field._metadata = field.metadata().override(paramId=shortname_to_paramid(newname))

    return soil_data


def regridding(
    fields: ekd.FieldList, grid: Optional[Union[str, list[float]]], area: Optional[list[float]], template: ekd.Field
) -> ekd.FieldList:
    """Apply regridding to the field.

    Parameters
    ----------
    fields : ekd.FieldList
        Fields to be regridded.
    grid : Optional[Union[str, list[float]]]
        Grid for the regridding.
    area : Optional[list[float]]
        Area for the regridding.
    template : ekd.Field
        Template for the regridding.

    Returns
    -------
    ekd.FieldList
        Regridded fields.
    """
    import earthkit.regrid as ekr
    import numpy as np

    r = ekd.FieldList()

    _ = area

    f_md = template.metadata()

    for f in fields:
        rolled_values = np.roll(f.to_numpy(), -f.shape[1] // 2, axis=1)
        interpolated_values = ekr.interpolate(rolled_values, in_grid={"grid": (0.25, 0.25)}, out_grid={"grid": grid})

        # Set the metadata with the grid directly, not this ridiculous template, TODO Harrison Cook
        namespace_metadata = f.metadata().as_namespace("mars")
        namespace_metadata.update(f.metadata().as_namespace("time"))
        namespace_metadata["paramId"] = shortname_to_paramid(namespace_metadata.pop("param"))

        for k in ["typeOfLevel", "time", "date"]:
            namespace_metadata[k] = f.metadata()[k]
        for k in ["domain", "levtype", "type", "step", "validityDate", "validityTime", "timespan"]:
            namespace_metadata.pop(k, None)

        r += r.from_array(np.expand_dims(interpolated_values, 0), f_md.override(**namespace_metadata))  # type: ignore
    return r


def retrieve(
    requests: list[dict[str, Any]],
    grid: Optional[Union[str, list[float]]],
    area: Optional[list[float]],
    template_provider: TemplateProvider,
    patch: Optional[Any] = None,
    **kwargs: Any,
) -> ekd.FieldList:
    """Retrieve data from ECMWF Opendata.

    Parameters
    ----------
    requests : list[dict[str, Any]]
        The list of requests to be retrieved.
    grid : Optional[Union[str, list[float]]]
        The grid for the retrieval.
    area : Optional[list[float]]
        The area for the retrieval.
    template_provider : TemplateProvider
        Template provider for the regridding.
    patch : Optional[Any], optional
        Optional patch for the request, by default None.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    ekd.FieldList
        The retrieved data.
    """

    def _(r: DataRequest):
        mars = r.copy()
        for k, v in r.items():
            if isinstance(v, (list, tuple)):
                mars[k] = "/".join(str(x) for x in v)
            else:
                mars[k] = str(v)

        return ",".join(f"{k}={v}" for k, v in mars.items())

    result = ekd.SimpleFieldList()
    for r in requests:
        r.update(kwargs)

        if patch:
            r = patch(r)

        template = template_provider.template(None, {"grid": grid, "levtype": r["levtype"]})  # type: ignore

        if any(k in r["param"] for k in SOIL_MAPPING.keys()):
            requested_soil_variables = [k for k in SOIL_MAPPING.keys() if k in r["param"]]
            r["param"] = [p for p in r["param"] if p not in requested_soil_variables]
            result += regridding(_retrieve_soil(r, requested_soil_variables), grid, area, template)

        LOG.debug("%s", _(r))
        result += regridding(ekd.from_source("ecmwf-open-data", r), grid, area, template)  # type: ignore

    return result


class OpenDataInputPlugin(GribInput):
    """Get input fields from ECMWF open-data."""

    trace_name = "opendata"

    def __init__(
        self,
        context: Context,
        *,
        pre_processors: Optional[list[ProcessorConfig]] = None,
        templates: Optional[Union[str, dict]] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the OpenDataInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        namer : Optional[Any]
            Optional namer for the input.
        templates : Optional[Union[str, dict]], optional
            Optional specification for template provider, by default None
        """
        rules_for_namer = [
            ({"levtype": "sol"}, "{param}"),
        ]
        super().__init__(context, namer={"rules": rules_for_namer}, pre_processors=pre_processors)
        self.pre_processors.append(OrographyProcessor(context=context, orog="gh"))

        self.variables = self.checkpoint.variables_from_input(include_forcings=False)
        self.kwargs = kwargs

        self.template_provider = create_template_provider(self, templates or "builtin")  # type: ignore

    def create_input_state(self, *, date: Optional[Date]) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """
        if date is None:
            date = to_datetime(-1)
            LOG.warning("OpenDataInput: `date` parameter not provided, using yesterday's date: %s", date)

        date = to_datetime(date)

        return self._create_input_state(
            self.retrieve(
                self.variables,
                [date + h for h in self.checkpoint.lagged],
            ),
            variables=self.variables,
            date=date,
        )

    def retrieve(self, variables: list[str], dates: list[Date]) -> Any:
        """Retrieve data for the given variables and dates.

        Parameters
        ----------
        variables : list[str]
            The list of variables to retrieve.
        dates : list[Any]
            The list of dates for which to retrieve the data.

        Returns
        -------
        Any
            The retrieved data.
        """

        requests = self.checkpoint.mars_requests(
            variables=variables,
            dates=dates,
            use_grib_paramid=self.context.use_grib_paramid,
            type="fc",
        )

        if not requests:
            raise ValueError("No requests for %s (%s)" % (variables, dates))

        kwargs = self.kwargs.copy()

        return retrieve(
            requests,
            self.checkpoint.grid,
            self.checkpoint.area,
            template_provider=self.template_provider,
            patch=self.patch_data_request,
            **kwargs,
        )

    def load_forcings_state(self, *, variables: list[str], dates: list[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            The list of variables for which to load the forcings state.
        dates : list[Date]
            The list of dates for which to load the forcings state.
        current_state : State
            The current state to be updated with the loaded forcings state.

        Returns
        -------
        Any
            The loaded forcings state.
        """
        return self._load_forcings_state(
            self.retrieve(variables, dates), variables=variables, dates=dates, current_state=current_state
        )



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import TYPE_CHECKING

import earthkit.data as ekd
import tqdm
from anemoi.inference.context import Context
from anemoi.inference.processor import Processor

if TYPE_CHECKING:
    from earthkit.data.readers.grib.codes import GribField

LOG = logging.getLogger(__name__)


def _mir_regrid(field: "GribField", grid, area) -> "GribField":
    import io

    import mir
    from earthkit.data import create_encoder

    encoder = create_encoder("grib")
    message = encoder.encode(field).to_bytes()

    mir_input = mir.GribMemoryInput(message)  # type: ignore
    job_args = {"grid": grid}
    if area:
        job_args["area"] = area

    job = mir.Job(**job_args)  # type: ignore
    buffer = io.BytesIO()

    job.execute(mir_input, buffer)

    return ekd.from_source("memory", buffer.getvalue())[0]  # type: ignore


def regrid(fields: ekd.FieldList, grid, area) -> ekd.FieldList:
    """Regrid a list of fields to a specified grid and area.

    TO BE REPLACED WITH EARTHKIT-REGRID
    """
    if isinstance(grid, (list, tuple)):
        grid = "/".join(map(str, grid))
    if isinstance(area, (list, tuple)):
        area = "/".join(map(str, area))

    result = list(map(lambda f: _mir_regrid(f, grid, area), tqdm.tqdm(fields, desc="Regridding fields")))  # type: ignore
    return ekd.FieldList.from_fields(result)


class RegridPreprocessor(Processor):
    """Regrid an input fieldlist.

    Can only be used when working with grib, from any of the earthkit data sources.
    i.e. mars, cds, opendata, grib files, etc.
    """

    def __init__(self, context: Context, grid: str | list[float], area: str | list[float] | None = None) -> None:
        """Initialize the Regridding processor.

        Parameters
        ----------
        context : Context
            The context in which the processor operates.
        grid : str | list[float]
            The target grid for regridding.
        area : str | list[float] | None, optional
            The target area for regridding, by default None
        """
        super().__init__(context)
        self._grid = grid
        self._area = area

    def process(self, fields: ekd.FieldList) -> ekd.FieldList:  # type: ignore
        """Process the fields to replace NaNs with the mean value.

        Parameters
        ----------
        fields : list
            List of fields to process.

        Returns
        -------
        list
            List of processed fields with NaNs replaced by the mean value.
        """
        return regrid(fields, self._grid, self._area)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from anemoi.inference.context import Context
from anemoi.inference.decorators import main_argument
from anemoi.inference.inputs.mars import MarsInput
from anemoi.inference.inputs.mars import postproc
from anemoi.inference.types import DataRequest
from anemoi.inference.types import Date

LOG = logging.getLogger(__name__)


def retrieve(
    collection: str,
    requests: List[Dict[str, Any]],
    grid: Optional[Union[str, List[float]]],
    area: Optional[List[float]],
    patch: Optional[Any] = None,
    **kwargs: Any,
) -> Any:
    """Retrieve data from Polytope.

    Parameters
    ----------
    collection : str
        The polytope collection to use for retrieval.
    requests : List[Dict[str, Any]]
        The list of requests to be retrieved.
    grid : Optional[Union[str, List[float]]]
        The grid for the retrieval.
    area : Optional[List[float]]
        The area for the retrieval.
    patch : Optional[Any], optional
        Optional patch for the request, by default None.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    Any
        The retrieved data.
    """
    import earthkit.data as ekd

    def _(r: DataRequest) -> str:
        mars = r.copy()
        for k, v in r.items():
            if isinstance(v, (list, tuple)):
                mars[k] = "/".join(str(x) for x in v)
            else:
                mars[k] = str(v)

        return ",".join(f"{k}={v}" for k, v in mars.items())

    pproc = postproc(grid, area)

    result = ekd.from_source("empty")
    for r in requests:
        if r.get("class") in ("rd", "ea"):
            r["class"] = "od"

        # ECMWF operational data has stream oper for 00 and 12 UTC and scda for 06 and 18 UTC

        if r.get("type") == "fc" and r.get("stream") == "oper" and r["time"] in ("0600", "1800"):
            r["stream"] = "scda"

        r.update(pproc)
        r.update(kwargs)

        if patch:
            r = patch(r)

        LOG.debug("%s", _(r))

        result += ekd.from_source("polytope", collection, r, stream=False)

    return result


@main_argument("collection")
class PolytopeInputPlugin(MarsInput):
    """Get input fields from Polytope."""

    trace_name = "polytope"

    def __init__(
        self,
        context: Context,
        collection: Optional[str] = None,
        **kwargs: Any,
    ):
        """Initialize the Polytope input plugin.

        Parameters
        ----------
        context : Context
            The context for the input plugin.
        collection : Optional[str], optional
            The collection to use for retrieval, by default 'ecmwf-mars'.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, **kwargs)
        self.collection = collection or "ecmwf-mars"

    def retrieve(self, variables: List[str], dates: List[Date]) -> Any:
        """Retrieve data for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            The list of variables to retrieve.
        dates : List[Any]
            The list of dates for which to retrieve the data.

        Returns
        -------
        Any
            The retrieved data.
        """
        requests = self.checkpoint.mars_requests(
            variables=variables,
            dates=dates,
            use_grib_paramid=self.context.use_grib_paramid,
            patch_request=self.context.patch_data_request,
        )

        if not requests:
            raise ValueError("No requests for %s (%s)" % (variables, dates))

        kwargs = self.kwargs.copy()
        kwargs.setdefault("expver", "0001")

        return retrieve(
            self.collection,
            requests,
            self.checkpoint.grid,
            self.checkpoint.area,
            self.patch,
            **kwargs,
        )



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import patch

from anemoi.inference.runners import create_runner
from anemoi.inference.testing import fake_checkpoints
from anemoi.inference.testing.mock_checkpoint import MockRunConfiguration
from anemoi.plugins.ecmwf.inference.multio import MultioOutputPlugin


@patch("anemoi.plugins.ecmwf.inference.multio.MultioOutputPlugin.open")
@fake_checkpoints
def test_metadata(mock_open_function) -> None:
    """Test the inference process using a fake checkpoint.

    This function loads a configuration, creates a runner, and runs the inference
    process to ensure that the system works as expected with the provided configuration.
    """
    # Mock the _server property
    mocked_server = MagicMock()
    mocked_write_field = MagicMock()
    mocked_flush = MagicMock()

    mocked_server.attach_mock(mocked_write_field, "write_field")
    mocked_server.attach_mock(mocked_flush, "flush")

    mock_open_function.side_effect = lambda state: setattr(MultioOutputPlugin, "_server", mocked_server)

    # Load configuration
    config = MockRunConfiguration.load(
        (Path(__file__).parent / "configs/multio.yaml").absolute(),
        overrides=dict(runner="testing", device="cpu", input="dummy"),
    )

    # Create runner and execute
    runner = create_runner(config)

    # Check calls to the _server property
    assert hasattr(runner.create_output(), "_server")

    runner.execute()

    mocked_write_field.assert_called()
    mocked_flush.assert_called()



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import Any

import pytest
from anemoi.plugins.ecmwf.inference.mir_templates import MirTemplatesProvider


class MockTemplateManager:
    pass


VALID_LOOKUPS = [
    {"levtype": "pl"},
    {"levtype": "sfc"},
]


@pytest.fixture
def MirTemplatesProvider_instance():
    """Fixture to create an instance of MirTemplatesProvider."""
    manager = MockTemplateManager()
    return MirTemplatesProvider(manager)


@pytest.mark.parametrize("grid", ["O48", "O96", "N320", "O1280", "1.0/1.0", "0.5/0.5", "0.25/0.25", [0.5, 0.5]])
@pytest.mark.parametrize("lookup", VALID_LOOKUPS)
def test_simple_regrid(MirTemplatesProvider_instance: MirTemplatesProvider, lookup: dict[str, Any], grid: str):
    """Test that the simple regrid template is correctly done."""
    provider = MirTemplatesProvider_instance
    variable = "lsm"

    lookup.update(grid=grid)

    template = provider.template(variable, lookup)

    assert template is not None, f"Template for {variable} with grid {grid} not found"
    expected_grid = grid if "/" not in grid else list(map(float, grid.split("/")))

    assert (
        template.metadata().geography.mars_grid() == expected_grid
    ), f"Expected grid {expected_grid}, got {template.metadata().geography.mars_grid()}"


@pytest.mark.parametrize("grid", ["0.5/0.5", "0.25/0.25", [0.5, 0.5]])
@pytest.mark.parametrize(
    "area_box",
    [
        ("10/-0.5/0.5/11", [10.0, 359.5, 0.5, 11.0]),
        ("2/4/2/4", [2.0, 4.0, 2.0, 4.0]),
        ([2, 4, 2, 4], [2.0, 4.0, 2.0, 4.0]),
    ],
)
@pytest.mark.parametrize("lookup", VALID_LOOKUPS)
def test_regrid_and_area(
    MirTemplatesProvider_instance: MirTemplatesProvider,
    lookup: dict[str, Any],
    grid: str,
    area_box: tuple[str, tuple[int, ...]],
):
    """Test that the regrid and area are correctly applied."""
    provider = MirTemplatesProvider_instance
    variable = "lsm"

    area, box = area_box
    lookup.update(grid=grid, area=area)

    template = provider.template(variable, lookup)

    assert template is not None, f"Template for {variable} with grid {grid} not found"
    expected_grid = grid if "/" not in grid else list(map(float, grid.split("/")))

    assert (
        template.metadata().geography.mars_grid() == expected_grid
    ), f"Expected grid {expected_grid}, got {template.metadata().geography.mars_grid()}"
    assert (
        template.metadata().geography.mars_area() == box
    ), f"Expected box {box}, got {template.metadata().geography.mars_area()}"



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from pathlib import Path

import pytest
from anemoi.inference.outputs import create_output
from anemoi.inference.runners import create_runner
from anemoi.inference.testing import fake_checkpoints
from anemoi.inference.testing.mock_checkpoint import MockRunConfiguration

FAKE_METADATA_KEYS = {
    "stream": "test",
    "expver": "1",
    "klass": "test",
    "type": "test",
    "model": "test",
}


@pytest.mark.parametrize(
    "plugin, kwargs",
    [
        ("multio", {"plan": None}),
        ("multio.grib", {"path": __file__}),
        ("multio.fdb", {"fdb_config": {}}),
    ],
)
@fake_checkpoints
def test_plugin(plugin, kwargs):
    config = MockRunConfiguration.load(
        (Path(__file__).parent / "configs/simple.yaml").absolute(),
        overrides=dict(output={plugin: {**kwargs, **FAKE_METADATA_KEYS}}),
    )
    runner = create_runner(config)
    output = create_output(runner, config.output)
    assert output is not None


if __name__ == "__main__":
    test_plugin()
